/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContainer;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeID;
import org.deckfour.xes.model.XAttributeList;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XVisitor;
import org.deckfour.xes.util.XAttributeUtils;

/**
 * @author F. Mannhardt
 * 
 */
abstract class XAttributeExternalImpl implements XAttribute, ExternalAttributable {

	private static final long serialVersionUID = 1L;

	protected static final String IMMUTABLE_ERROR = "This XAttribute is immutable, as it may be serialized by the storage engine at any time!";

	private int key = Integer.MIN_VALUE;
	private XExtension extension;

	// Support for meta-attributes
	private ExternalStore store;
	private long externalId = Long.MIN_VALUE;

	protected XAttributeExternalImpl(int key, ExternalStore store) {
		this(key, null, store);
	}

	protected XAttributeExternalImpl(int key, XExtension extension, ExternalStore store) {
		this.key = key;
		this.extension = extension;
		this.store = store;
	}

	public abstract Object clone();

	int getInternalKey() {
		return key;
	}

	@Override
	public String getKey() {
		assert key != Integer.MIN_VALUE : "Internal key of attribute is not initialized properly!";
		String s = store.getAttributeKeyPool().getValue(key);
		assert s != null : "Key " + key + " of attribute is unknown in pool: " + store.getAttributeKeyPool().toString();
		return s;
	}

	@Override
	public XExtension getExtension() {
		return extension;
	}

	@Override
	public boolean hasAttributes() {
		return externalId != Long.MIN_VALUE ? true : false;
	}

	@Override
	public synchronized XAttributeMap getAttributes() {
		if (hasAttributes()) {
			return getStore().getAttributes(this);
		} else {
			this.externalId = store.getIdFactory().nextId();
			return getStore().getAttributes(this);
		}
	}

	@Override
	public synchronized void setAttributes(XAttributeMap attributes) {
		if (hasAttributes()) {
			store.setAttributes(this, attributes);
		} else {
			this.externalId = store.getIdFactory().nextId();
			store.setAttributes(this, attributes);
		}
	}

	@Override
	public Set<XExtension> getExtensions() {
		if (hasAttributes()) {
			Set<XExtension> metaExtensions = XAttributeUtils.extractExtensions(getAttributes());
			Set<XExtension> allExtensions = new HashSet<>(metaExtensions);
			if (getExtension() != null) {
				allExtensions.add(getExtension());	
			}			
			return allExtensions;	
		} else {
			return Collections.emptySet();
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof XAttributeExternalImpl) {
			return ((XAttributeExternalImpl) obj).key == key;
		} else if (obj instanceof XAttribute) {
			XAttribute other = (XAttribute) obj;
			return other.getKey().equals(getKey());
		} else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return getKey().hashCode();
	}

	@Override
	public int compareTo(XAttribute o) {
		if (o instanceof XAttributeExternalImpl) {
			return compare(key, ((XAttributeExternalImpl) o).key);
		} else {
			return getKey().compareTo(o.getKey());
		}
	}

	private static int compare(int x, int y) {
		return (x < y) ? -1 : ((x == y) ? 0 : 1);
	}

	@Override
	public void accept(XVisitor visitor, XAttributable parent) {
		visitor.visitAttributePre(this, parent);
		// Does not do anything here!
		visitor.visitAttributePost(this, parent);
	}

	public void setExtension(XExtension extension) {
		this.extension = extension;
	}

	@Override
	public long getExternalId() {
		return externalId;
	}

	void setExternalId(long id) {
		this.externalId = id;
	}

	private ExternalStore getStore() {
		return store;
	}

	private void setStore(ExternalStore store) {
		this.store = store;
	}

	private void setKey(int key) {
		this.key = key;
	}

	static XAttributeExternalImpl convert(ExternalStore store, XAttribute value) {
		final StringPool stringPool = store.getAttributeKeyPool();
		if (value instanceof XAttributeExternalImpl) {
			XAttributeExternalImpl externalAttr = (XAttributeExternalImpl) value;
			if (externalAttr.getStore() == store) {
				// This attribute belongs to the same store, so no need to do anything
				return externalAttr;
			} else {
				// This attribute belongs to a different store, we to create a new instance related to our store
				return convertStandardAttributes(store, externalAttr, stringPool);
			}
		}
		return convertStandardAttributes(store, value, stringPool);
	}

	private static XAttributeExternalImpl convertStandardAttributes(ExternalStore store, XAttribute value,
			StringPool stringPool) {
		int internalAttributeKey = stringPool.put(value.getKey());
		if (value instanceof XAttributeLiteral) {
			return new XAttributeLiteralExternalImpl(internalAttributeKey, ((XAttributeLiteral) value).getValue(),
					value.getExtension(), store);
		} else if (value instanceof XAttributeDiscrete) {
			return new XAttributeDiscreteExternalImpl(internalAttributeKey, ((XAttributeDiscrete) value).getValue(),
					value.getExtension(), store);
		} else if (value instanceof XAttributeContinuous) {
			return new XAttributeContinuousExternalImpl(internalAttributeKey,
					((XAttributeContinuous) value).getValue(), value.getExtension(), store);
		} else if (value instanceof XAttributeBoolean) {
			return new XAttributeBooleanExternalImpl(internalAttributeKey, ((XAttributeBoolean) value).getValue(),
					value.getExtension(), store);
		} else if (value instanceof XAttributeTimestamp) {
			return new XAttributeTimestampExternalImpl(internalAttributeKey, ((XAttributeTimestamp) value).getValue(),
					value.getExtension(), store);
		} else if (value instanceof XAttributeID) {
			return new XAttributeIDExternalImpl(internalAttributeKey, ((XAttributeID) value).getValue(),
					value.getExtension(), store);
		} else if (value instanceof XAttributeContainer) {
			return new XAttributeContainerExternalImpl(internalAttributeKey, value.getExtension(), store);
		} else if (value instanceof XAttributeList) {
			return new XAttributeListExternalImpl(internalAttributeKey, value.getExtension(), store);
		} else {
			throw new IllegalArgumentException("Attribute type unsupported by XESLite "
					+ value.getClass().getSimpleName());
		}
	}

	static XAttributeExternalImpl decorate(XAttributeExternalImpl undecoratedAttr, int key, MapDBStore store) {
		if (undecoratedAttr == null) {
			return null;
		}
		undecoratedAttr.setKey(key);
		undecoratedAttr.setStore(store);
		return undecoratedAttr;
	}

}